# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

# This is the exported file from the library when installed as a package in other projects. Whatever is imported in
# this file is accessible in projects that install `fy_library` package. Currently, CLI tool uses `Main_Flow` to run
# the tool.

from fy_library.fy import fy

__all__ = ["fy"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow create_property_template_model__using_parsed_fy_py_file -> PropertyTemplateModel:
    property parsed_fy_py_file using setter
    property abstract_entities_ordering_index using setter
    property abstract_mixins using parsed_property_fy_py_file
    property mro_ordered_abstract_mixins using abstract_mixins_and_ordered_abstract_entities
fy"""

from typing import Any, Dict

from fy_core.base.flow_base import FlowBase
from fy_library.domain.fy_py_template_models import PropertyTemplateModel
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile, ParsedPropertyFyPyFile
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)

from fy_library.mixins.property.abstract_mixins.using_parsed_property_fy_py_file_fy import (
    AbstractMixins_UsingParsedPropertyFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.mro_ordered_abstract_mixins.new_parsed_fy_py_file_and_ordered_abstract_entities_fy import (
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
)

from fy_library.mixins.property.ordered_abstract_entities.using_setter import (
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class CreatePropertyTemplateModel_UsingParsedFyPyFile_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
    AbstractMixins_UsingParsedPropertyFyPyFile_PropertyMixin,
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
    # Base
    FlowBase[PropertyTemplateModel],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        abstract_entities_ordering_index: Dict[str, int],
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        self._abstract_entities_ordering_index = abstract_entities_ordering_index
        super().__init__(*args, **kwargs)

    def __call__(self) -> PropertyTemplateModel:
        # fy:end <<<===
        parsed_property_fy_py_file = self._parsed_fy_py_file
        assert isinstance(parsed_property_fy_py_file, ParsedPropertyFyPyFile)
        return PropertyTemplateModel(
            python_class_name=parsed_property_fy_py_file.python_class_name,
            property_name=parsed_property_fy_py_file.property_name,
            implementation_name=parsed_property_fy_py_file.implementation_name,
            abstract_property_mixins=self._mro_ordered_abstract_mixins,
            generics_def=parsed_property_fy_py_file.generics_def,
            property_type=parsed_property_fy_py_file.property_type,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.mixin_models import MethodMixinModel


property optional_method_mixin_model: MethodMixinModel | None using mixin_line:
    property mixin_line
fy"""

import abc
import re
from functools import cached_property
from typing import Final

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
)
from fy_library.domain.mixin_models import MethodMixinModel, MixinModelKind
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.mixin_line.abc_fy import (
    MixinLine_PropertyMixin_ABC,
)
from fy_library.mixins.property.optional_method_mixin_model.abc_fy import (
    OptionalMethodMixinModel_PropertyMixin_ABC,
)

_FLOW_METHOD_REGEX: Final = re.compile(
    rf"^\s+method\s+(?P<method_name>{FY_ENTITY_REGEX_STRING})\s+"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*"
    rf"(?:\[(?P<generics_impl>{PYTHON_MULTI_ENTITY_REGEX_STRING})])?$"
)


# fy:start ===>>>
class OptionalMethodMixinModel_UsingMixinLine_PropertyMixin(
    # Property_mixins
    MixinLine_PropertyMixin_ABC,
    OptionalMethodMixinModel_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _optional_method_mixin_model(self) -> MethodMixinModel | None:
        # fy:end <<<===
        flow_method_fy_search = _FLOW_METHOD_REGEX.search(self._mixin_line)

        if flow_method_fy_search is None:
            return None

        method_name = PythonEntityName.from_snake_case(
            flow_method_fy_search.group("method_name")
        )
        implementation_name = PythonEntityName.from_snake_case(
            flow_method_fy_search.group("implementation_name")
        )
        return MethodMixinModel(
            python_class_name=PythonEntityName.from_pascal_case(
                f"{ method_name.pascal_case }_Using{ implementation_name.pascal_case }_MethodMixin"
            ),
            kind=MixinModelKind.METHOD,
            method_name=method_name,
            implementation_name=implementation_name,
            generics_impl=flow_method_fy_search.group("generics_impl") or "",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.mixins.property.base_flow_file_split.abc_fy import BaseFlowFileSplitModel


property base_flow_file_split: BaseFlowFileSplitModel using base_flow_regex:
    property fy_py_file_to_parse
    property fy_code
fy"""

import abc
import re
from functools import cached_property
from typing import Final, List

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
)
from fy_library.domain.annotation_object import Annotation, AnnotationKind
from fy_library.mixins.property.base_flow_file_split.abc_fy import (
    BaseFlowFileSplitModel,
    BaseFlowFileSplit_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)


_BASE_FLOW_STRING_SPLIT_REGEX: Final = re.compile(
    rf"base\s+flow\s+(?P<flow_name>{FY_ENTITY_REGEX_STRING})\s*"
    rf"(?:\[(?P<generics_def>{PYTHON_MULTI_ENTITY_REGEX_STRING})])?\s*"
    rf"(?:\((?P<declared_base_flow>{FY_ENTITY_REGEX_STRING})\))?"
    rf"(?:\s+->\s+(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING}))?:\s*\n"
)

_CHECK_ANNOTATIONS: Final = re.compile(r"(?P<annotations>@\w+)")


# fy:start ===>>>
class BaseFlowFileSplit_UsingBaseFlowRegex_PropertyMixin(
    # Property_mixins
    BaseFlowFileSplit_PropertyMixin_ABC,
    FyCode_PropertyMixin_ABC,
    FyPyFileToParse_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _base_flow_file_split(self) -> BaseFlowFileSplitModel:
        # fy:end <<<===
        base_flow_file_split = _BASE_FLOW_STRING_SPLIT_REGEX.split(self._fy_code)

        check_if_annotations_are_ordered = [
            bool(re.match(_CHECK_ANNOTATIONS, maybe_annotation))
            for maybe_annotation in base_flow_file_split[0].split("\n")
            if maybe_annotation != ""
        ]

        first_annotation_found: bool = False
        for is_annotation in check_if_annotations_are_ordered:
            if is_annotation and not first_annotation_found:
                first_annotation_found = True
            if not is_annotation and first_annotation_found:
                raise AssertionError(
                    f"Base flow has unordered annotations in {base_flow_file_split[1]}_BaseFlow."
                )

        annotations: List[Annotation] = []

        all_annotations = re.findall(_CHECK_ANNOTATIONS, base_flow_file_split[0])
        for annotation in all_annotations:
            annotation_object = Annotation(
                kind=AnnotationKind(annotation),
            )
            annotations.append(annotation_object)

        user_imports = re.sub(_CHECK_ANNOTATIONS, "", base_flow_file_split[0])

        assert (
            (len(base_flow_file_split)) == 6
        ), f"Flow file split length {len(base_flow_file_split)} is invalid in {self._fy_py_file_to_parse}"

        base_flow_file_split_model = BaseFlowFileSplitModel(
            user_imports=user_imports,
            annotations=annotations,
            base_flow_name=base_flow_file_split[1],
            generics_def=base_flow_file_split[2] or "",
            declared_base_flow=base_flow_file_split[3] or "",
            return_type=base_flow_file_split[4],
            mixins=base_flow_file_split[5],
        )

        return base_flow_file_split_model



# fy:start ===>>>
from typing import Dict


class AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin:
    @property
    def _abstract_entities_ordering_index(self) -> Dict[str, int]:
        return self.__abstract_entities_ordering_index

    @_abstract_entities_ordering_index.setter
    def _abstract_entities_ordering_index(
        self, abstract_entities_ordering_index: Dict[str, int]
    ) -> None:
        self.__abstract_entities_ordering_index = abstract_entities_ordering_index


# fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List
from fy_library.domain.mixin_models import PropertyMixinModel


property property_constant_setter_mixins: List[PropertyMixinModel] using property_mixins:
    property property_mixins
fy"""

import abc
from functools import cached_property

from fy_library.constants import PROPERTY_CONSTANT_IMPLEMENTATION_NAME
from fy_library.mixins.property.property_constant_setter_mixins.abc_fy import (
    PropertyConstantSetterMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.property_mixins.abc_fy import (
    PropertyMixins_PropertyMixin_ABC,
)
from fy_library.domain.mixin_models import PropertyMixinModel
from typing import List


# fy:start ===>>>
class PropertyConstantSetterMixins_UsingPropertyMixins_PropertyMixin(
    # Property_mixins
    PropertyConstantSetterMixins_PropertyMixin_ABC,
    PropertyMixins_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _property_constant_setter_mixins(self) -> List[PropertyMixinModel]:
        # fy:end <<<===
        return [
            property_mixin
            for property_mixin in self._property_mixins
            if property_mixin.implementation_name.snake_case
            == PROPERTY_CONSTANT_IMPLEMENTATION_NAME
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import Dict


property abstract_entities_ordering_index: Dict[str, int]
fy"""

import abc

from typing import Dict


# fy:start ===>>>
class AbstractEntitiesOrderingIndex_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _abstract_entities_ordering_index(self) -> Dict[str, int]:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import Dict


property abstract_entities_ordering_index: Dict[str, int] using parsed_fy_py_files_map_by_key:
    property parsed_fy_py_files_map_by_key
fy"""

from functools import cached_property

from typing import Dict

import abc

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)

from fy_library.mixins.property.ordered_abstract_entities.abc_fy import (
    AbstractEntitiesOrderingIndex_PropertyMixin_ABC,
)


# fy:start ===>>>
class AbstractEntitiesOrderingIndex_UsingParsedFyPyFilesMapByKey_PropertyMixin(
    # Property_mixins
    AbstractEntitiesOrderingIndex_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _abstract_entities_ordering_index(self) -> Dict[str, int]:
        # fy:end <<<===
        abstract_entities_ordering_index = {
            entity_key: entity_num
            for entity_num, entity_key in enumerate(
                sorted(self._parsed_fy_py_files_map_by_key)
            )
            if self._parsed_fy_py_files_map_by_key[entity_key].file_type
            in {
                ParsedFyPyFileKind.ABSTRACT_METHOD,
                ParsedFyPyFileKind.ABSTRACT_PROPERTY,
            }
        }
        return abstract_entities_ordering_index



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow create_abstract_method_template_model__using_parsed_fy_py_file -> AbstractMethodTemplateModel:
    property parsed_fy_py_file using setter
fy"""

from typing import Any

from fy_core.base.flow_base import FlowBase

from fy_library.domain.fy_py_template_models import AbstractMethodTemplateModel
from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFile,
    ParsedAbstractMethodFyPyFile,
)
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class CreateAbstractMethodTemplateModel_UsingParsedFyPyFile_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    # Base
    FlowBase[AbstractMethodTemplateModel],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        super().__init__(*args, **kwargs)

    def __call__(self) -> AbstractMethodTemplateModel:
        # fy:end <<<===
        parsed_abstract_method_fy_py_file = self._parsed_fy_py_file

        assert isinstance(
            parsed_abstract_method_fy_py_file, ParsedAbstractMethodFyPyFile
        )
        return AbstractMethodTemplateModel(
            python_class_name=parsed_abstract_method_fy_py_file.python_class_name,
            abstract_method_name=parsed_abstract_method_fy_py_file.abstract_method_name,
            generics_def=parsed_abstract_method_fy_py_file.generics_def,
            arguments=parsed_abstract_method_fy_py_file.arguments,
            return_type=parsed_abstract_method_fy_py_file.return_type,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import Dict
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


property parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile] using parsed_fy_py_files:
    property parsed_fy_py_files
fy"""

import abc
from functools import cached_property
from typing import Dict

from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.parsed_fy_py_files.abc_fy import (
    ParsedFyPyFiles_PropertyMixin_ABC,
)

from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)


# fy:start ===>>>
class ParsedFyPyFilesMapByKey_UsingParsedFyPyFiles_PropertyMixin(
    # Property_mixins
    ParsedFyPyFiles_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _parsed_fy_py_files_map_by_key(self) -> Dict[str, ParsedFyPyFile]:
        # fy:end <<<===
        parsed_fy_py_files_map_by_key: dict[str, ParsedFyPyFile] = {}
        for parsed_fy_py_file in self._parsed_fy_py_files:
            if parsed_fy_py_file.entity_key in parsed_fy_py_files_map_by_key:
                raise AssertionError(
                    f"Duplicate key {parsed_fy_py_file.entity_key} found."
                )
            parsed_fy_py_files_map_by_key[parsed_fy_py_file.entity_key] = (
                parsed_fy_py_file
            )

        return parsed_fy_py_files_map_by_key



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind


property fy_file_kind: ParsedFyPyFileKind using fy_code:
    property fy_py_file_to_parse
    property fy_code
fy"""

import abc
import re
from functools import cached_property
from typing import Final

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
    PYTHON_ARGUMENTS_REGEX_STRING,
)
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFileKind
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_file_kind.abc_fy import (
    FyFileKind_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)

_FLOW_MATCH_REGEX: Final = re.compile(
    rf"^flow\s+{FY_ENTITY_REGEX_STRING}\s*(\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?\s*"
    rf"(\({FY_ENTITY_REGEX_STRING}\))?\s*"
    rf"(->\s*(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s*)?:\s*$",
)
_BASE_FLOW_MATCH_REGEX: Final = re.compile(
    rf"^base\s+flow\s+{FY_ENTITY_REGEX_STRING}\s*(\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?\s*"
    rf"(\({FY_ENTITY_REGEX_STRING}\))?\s*"
    rf"->\s*(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s*:\s*$",
)
_METHOD_MATCH_REGEX: Final = re.compile(
    rf"^method\s+(?P<method_name>{FY_ENTITY_REGEX_STRING})\s*"
    rf"(\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?"
    rf"(?P<arguments>\(({PYTHON_ARGUMENTS_REGEX_STRING})\))?\s+->"
    rf"\s+(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s+"
    rf"using\s+(?P<implementation_name>{FY_ENTITY_REGEX_STRING})\s*:\s*$"
)
_ABSTRACT_METHOD_REGEX: Final = re.compile(
    rf"^method\s+{FY_ENTITY_REGEX_STRING}\s*(\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?(\({PYTHON_ARGUMENTS_REGEX_STRING}\))?"
    rf"\s*|->\s*({PYTHON_MULTI_ENTITY_REGEX_STRING})\s*$",
)
_ABSTRACT_PROPERTY_REGEX: Final = re.compile(
    rf"^property\s+{FY_ENTITY_REGEX_STRING}\s*(\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?:\s*({PYTHON_MULTI_ENTITY_REGEX_STRING})\s*$",
)
_PROPERTY_REGEX: Final = re.compile(
    rf"^property\s+{FY_ENTITY_REGEX_STRING}\s*((\[{PYTHON_MULTI_ENTITY_REGEX_STRING}])?:\s*({PYTHON_MULTI_ENTITY_REGEX_STRING}))\s+using\s+{FY_ENTITY_REGEX_STRING}\s*:\s*$",
)


# fy:start ===>>>
class FyFileKind_UsingFyCode_PropertyMixin(
    # Property_mixins
    FyCode_PropertyMixin_ABC,
    FyFileKind_PropertyMixin_ABC,
    FyPyFileToParse_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _fy_file_kind(self) -> ParsedFyPyFileKind:
        # fy:end <<<===
        for fy_code_line in self._fy_code.split("\n"):
            if _FLOW_MATCH_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.FLOW
            if _BASE_FLOW_MATCH_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.BASE_FLOW
            if _METHOD_MATCH_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.METHOD
            if _ABSTRACT_METHOD_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.ABSTRACT_METHOD
            if _ABSTRACT_PROPERTY_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.ABSTRACT_PROPERTY
            if _PROPERTY_REGEX.match(fy_code_line) is not None:
                return ParsedFyPyFileKind.PROPERTY

        raise ValueError(f"Undetected file type for {self._fy_py_file_to_parse}")



# fy:start ===>>>
from typing import List
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


class ParsedFyPyFiles_UsingSetter_PropertyMixin:
    @property
    def _parsed_fy_py_files(self) -> List[ParsedFyPyFile]:
        return self.__parsed_fy_py_files

    @_parsed_fy_py_files.setter
    def _parsed_fy_py_files(self, parsed_fy_py_files: List[ParsedFyPyFile]) -> None:
        self.__parsed_fy_py_files = parsed_fy_py_files


# fy:end <<<===



# fy:start ===>>>
from typing import List


class CachedImport_UsingSetter_PropertyMixin:
    @property
    def _cached_import(self) -> List[str]:
        return self.__cached_import

    @_cached_import.setter
    def _cached_import(self, cached_import: List[str]) -> None:
        self.__cached_import = cached_import


# fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List



property import_base_flow: List[str] using declared_base_flow_name:
    property declared_base_flow_name
    property mixin_import_map
fy"""

from functools import cached_property

import abc
from typing import List

from fy_library.mixins.property.mixin_import_map.abc_fy import (
    MixinImportMap_PropertyMixin_ABC,
)

from fy_library.mixins.property.declared_base_flow_name.abc_fy import (
    DeclaredBaseFlowName_PropertyMixin_ABC,
)


# fy:start ===>>>
class ImportBaseFlow_UsingDeclaredBaseFlowName_PropertyMixin(
    # Property_mixins
    DeclaredBaseFlowName_PropertyMixin_ABC,
    MixinImportMap_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _import_base_flow(self) -> List[str]:
        # fy:end <<<===
        if self._declared_base_flow_name == "":
            return []
        return [self._mixin_import_map[self._declared_base_flow_name]]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property import_any: List[str] using property_setters_exists:
    property property_setter_mixins
    property property_constant_setter_mixins
fy"""

import abc
from functools import cached_property
from typing import List

from fy_library.mixins.property.property_setter_mixins.abc_fy import (
    PropertySetterMixins_PropertyMixin_ABC,
)

from fy_library.mixins.property.property_constant_setter_mixins.abc_fy import (
    PropertyConstantSetterMixins_PropertyMixin_ABC,
)


# fy:start ===>>>
class ImportAny_UsingPropertySettersExists_PropertyMixin(
    # Property_mixins
    PropertyConstantSetterMixins_PropertyMixin_ABC,
    PropertySetterMixins_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _import_any(self) -> List[str]:
        # fy:end <<<===
        return (
            ["from typing import Any"]
            if self._property_setter_mixins or self._property_constant_setter_mixins
            else []
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from typing import Final

FY_PY_FILE_SIGNATURE: Final = '"""fy\n'

FY_CODE_FILE_END_SIGNATURE: Final = 'fy"""'

FY_ENTITY_REGEX_STRING: Final = r"\w+"

PYTHON_ARGUMENTS_REGEX_STRING: Final = r"\s*[^)]*\s*"

_PYTHON_ENTITY_CHAR_REGEX_STRING: Final = r"[\w\[\]]"

PYTHON_MULTI_ENTITY_REGEX_STRING: Final = (
    rf"{_PYTHON_ENTITY_CHAR_REGEX_STRING}|"
    rf"{_PYTHON_ENTITY_CHAR_REGEX_STRING}[\w\.\[\]\t \|\,]*{_PYTHON_ENTITY_CHAR_REGEX_STRING}"
)

FY_START_MARKER: Final = "# fy:start ===>>>"

FY_END_MARKER: Final = "# fy:end <<<==="

FY_PY_FILE_EXTENSION: Final = "_fy.py"

PROPERTY_SETTER_IMPLEMENTATION_NAME: Final = "setter"

PROPERTY_CONSTANT_IMPLEMENTATION_NAME: Final = "constant"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property fy_py_file_content: str using required_property_setter:
    property parsed_fy_py_file
    property generated_fy_py_code
    property import_generic
fy"""

import abc
from functools import cached_property
from typing import Final

from fy_library.constants import FY_START_MARKER, FY_END_MARKER
from fy_library.domain.parsed_fy_py_file import PropertySetterFyPyFile
from fy_library.mixins.property.fy_py_file_content.abc_fy import (
    FyPyFileContent_PropertyMixin_ABC,
)
from fy_library.mixins.property.generated_fy_py_code.abc_fy import (
    GeneratedFyPyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)
from fy_library.mixins.property.import_generic.abc_fy import (
    ImportGeneric_PropertyMixin_ABC,
)

_NEW_LINE: Final = "\n"


# fy:start ===>>>
class FyPyFileContent_UsingRequiredPropertySetter_PropertyMixin(
    # Property_mixins
    FyPyFileContent_PropertyMixin_ABC,
    GeneratedFyPyCode_PropertyMixin_ABC,
    ImportGeneric_PropertyMixin_ABC,
    ParsedFyPyFile_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _fy_py_file_content(self) -> str:
        # fy:end <<<===
        assert isinstance(self._parsed_fy_py_file, PropertySetterFyPyFile)

        fy_py_file_content = (
            f"{FY_START_MARKER}\n"
            f"{self._import_generic[0] + _NEW_LINE if self._parsed_fy_py_file.generics_def != '' else ''}"
            f"{self._parsed_fy_py_file.user_imports}"
            f"{self._generated_fy_py_code}"
            f"{FY_END_MARKER}\n"
            f"{self._parsed_fy_py_file.post_marker_file_content}"
        )
        return fy_py_file_content



# fy:start ===>>>
from fy_library.domain.parsed_fy_py_file import ParsedFlowFyPyFile


class ParsedFlowFyPyFile_UsingSetter_PropertyMixin:
    @property
    def _parsed_flow_fy_py_file(self) -> ParsedFlowFyPyFile:
        return self.__parsed_flow_fy_py_file

    @_parsed_flow_fy_py_file.setter
    def _parsed_flow_fy_py_file(
        self, parsed_flow_fy_py_file: ParsedFlowFyPyFile
    ) -> None:
        self.__parsed_flow_fy_py_file = parsed_flow_fy_py_file


# fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property flow_file_split: FlowFileSplitModel
fy"""

import abc

from pydantic import BaseModel


class FlowFileSplitModel(BaseModel):
    user_imports: str
    flow_name: str
    generics_def: str
    declared_base_flow: str
    return_type: str
    mixins: str


# fy:start ===>>>
class FlowFileSplit_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _flow_file_split(self) -> FlowFileSplitModel:
        raise NotImplementedError()
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property mixin_imports: List[str] using parsed_fy_py_file:
    property parsed_fy_py_file
    property mixin_import_map
    property parsed_fy_py_files_map_by_key
fy"""

import abc
from functools import cached_property
from typing import List, cast

from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFileKind,
    ParsedMethodFyPyFile,
    ParsedFlowFyPyFile,
    ParsedPropertyFyPyFile,
    ParsedBaseFlowFyPyFile,
    ParsedAbstractPropertyFyPyFile,
    ParsedAbstractMethodFyPyFile,
)
from fy_library.flows.imports.abstract_method_imports_fy import (
    AbstractMethodImportsFlow_Flow,
)
from fy_library.flows.imports.abstract_property_imports_fy import (
    AbstractPropertyImportsFlow_Flow,
)
from fy_library.flows.imports.base_flow_imports_fy import BaseFlowImports_Flow
from fy_library.flows.imports.flow_imports_fy import FlowImports_Flow
from fy_library.flows.imports.method_imports_fy import MethodImports_Flow
from fy_library.flows.imports.property_imports_fy import PropertyImports_Flow
from fy_library.mixins.property.mixin_import_map.abc_fy import (
    MixinImportMap_PropertyMixin_ABC,
)
from fy_library.mixins.property.mixin_imports.abc_fy import (
    MixinImports_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)


# fy:start ===>>>
class MixinImports_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    MixinImportMap_PropertyMixin_ABC,
    MixinImports_PropertyMixin_ABC,
    ParsedFyPyFile_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _mixin_imports(self) -> List[str]:
        # fy:end <<<===
        match self._parsed_fy_py_file.file_type:
            case ParsedFyPyFileKind.FLOW:
                parsed_flow_fy_py_file = cast(
                    ParsedFlowFyPyFile, self._parsed_fy_py_file
                )
                return FlowImports_Flow(
                    property_mixins=parsed_flow_fy_py_file.properties,
                    parsed_fy_py_files_map_by_key=self._parsed_fy_py_files_map_by_key,
                    mixin_import_map=self._mixin_import_map,
                    method_mixins=cast(
                        ParsedFlowFyPyFile, self._parsed_fy_py_file
                    ).methods,
                    parsed_flow_fy_py_file=parsed_flow_fy_py_file,
                )()
            case ParsedFyPyFileKind.BASE_FLOW:
                parsed_base_flow_fy_py_file = cast(
                    ParsedBaseFlowFyPyFile, self._parsed_fy_py_file
                )
                return BaseFlowImports_Flow(
                    property_mixins=parsed_base_flow_fy_py_file.properties,
                    parsed_fy_py_files_map_by_key=self._parsed_fy_py_files_map_by_key,
                    mixin_import_map=self._mixin_import_map,
                    method_mixins=parsed_base_flow_fy_py_file.methods,
                    abstract_property_mixins=parsed_base_flow_fy_py_file.abstract_property_mixins,
                    abstract_method_mixins=parsed_base_flow_fy_py_file.abstract_method_mixins,
                    parsed_base_flow_fy_py_file=parsed_base_flow_fy_py_file,
                )()
            case ParsedFyPyFileKind.METHOD:
                parsed_method_fy_py_file = cast(
                    ParsedMethodFyPyFile, self._parsed_fy_py_file
                )
                return MethodImports_Flow(
                    abstract_property_mixins=parsed_method_fy_py_file.abstract_property_mixins,
                    abstract_method_mixins=parsed_method_fy_py_file.abstract_method_mixins,
                    mixin_import_map=self._mixin_import_map,
                    parsed_method_fy_py_file=parsed_method_fy_py_file,
                )()
            case ParsedFyPyFileKind.ABSTRACT_METHOD:
                return AbstractMethodImportsFlow_Flow(
                    parsed_abstract_method_fy_py_file=cast(
                        ParsedAbstractMethodFyPyFile, self._parsed_fy_py_file
                    ),
                )()
            case ParsedFyPyFileKind.ABSTRACT_PROPERTY:
                return AbstractPropertyImportsFlow_Flow(
                    parsed_abstract_property_fy_py_file=cast(
                        ParsedAbstractPropertyFyPyFile, self._parsed_fy_py_file
                    ),
                )()
            case ParsedFyPyFileKind.PROPERTY:
                parsed_property_fy_py_file = cast(
                    ParsedPropertyFyPyFile, self._parsed_fy_py_file
                )
                return PropertyImports_Flow(
                    abstract_property_mixins=parsed_property_fy_py_file.abstract_property_mixins,
                    mixin_import_map=self._mixin_import_map,
                    parsed_property_fy_py_file=parsed_property_fy_py_file,
                )()
            case _:
                raise NotImplementedError(
                    f"Mixin imports not implemented for file type '{self._parsed_fy_py_file.file_type}' in '{self._parsed_fy_py_file.file_path}'"
                )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property declared_base_flow_name: str using parsed_base_flow_fy_py_file:
    property parsed_base_flow_fy_py_file
fy"""

import abc
from functools import cached_property

from fy_library.mixins.property.parsed_base_flow_fy_py_file.abc_fy import (
    ParsedBaseFlowFyPyFile_PropertyMixin_ABC,
)

from fy_library.mixins.property.declared_base_flow_name.abc_fy import (
    DeclaredBaseFlowName_PropertyMixin_ABC,
)


# fy:start ===>>>
class DeclaredBaseFlowName_UsingParsedBaseFlowFyPyFile_PropertyMixin(
    # Property_mixins
    DeclaredBaseFlowName_PropertyMixin_ABC,
    ParsedBaseFlowFyPyFile_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _declared_base_flow_name(self) -> str:
        # fy:end <<<===
        return self._parsed_base_flow_fy_py_file.declared_base_flow



# fy:start ===>>>
from typing import List
from fy_library.domain.mixin_models import AbstractMethodModel


class AbstractMethodMixins_UsingSetter_PropertyMixin:
    @property
    def _abstract_method_mixins(self) -> List[AbstractMethodModel]:
        return self.__abstract_method_mixins

    @_abstract_method_mixins.setter
    def _abstract_method_mixins(
        self, abstract_method_mixins: List[AbstractMethodModel]
    ) -> None:
        self.__abstract_method_mixins = abstract_method_mixins


# fy:end <<<===



# fy:start ===>>>
from typing import Dict
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


class ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin:
    @property
    def _parsed_fy_py_files_map_by_key(self) -> Dict[str, ParsedFyPyFile]:
        return self.__parsed_fy_py_files_map_by_key

    @_parsed_fy_py_files_map_by_key.setter
    def _parsed_fy_py_files_map_by_key(
        self, parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile]
    ) -> None:
        self.__parsed_fy_py_files_map_by_key = parsed_fy_py_files_map_by_key


# fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import Dict


property mixin_import_map: Dict[str, str]
fy"""

from typing import Dict
import abc


# fy:start ===>>>
class MixinImportMap_PropertyMixin_ABC(abc.ABC):
    @property
    @abc.abstractmethod
    def _mixin_import_map(self) -> Dict[str, str]:
        raise NotImplementedError()
        # fy:end <<<===



# fy:start ===>>>
from typing import List


class MixinImports_UsingSetter_PropertyMixin:
    @property
    def _mixin_imports(self) -> List[str]:
        return self.__mixin_imports

    @_mixin_imports.setter
    def _mixin_imports(self, mixin_imports: List[str]) -> None:
        self.__mixin_imports = mixin_imports
        # fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


flow base_flow_imports -> List[str]:
    property property_mixins using setter
    property method_mixins using setter
    property abstract_property_mixins using setter
    property abstract_method_mixins using setter
    property parsed_fy_py_files_map_by_key using setter
    property mixin_import_map using setter
    property parsed_base_flow_fy_py_file using setter
    property import_generic using constant(["from typing import Generic"])
    property declared_base_flow_name using parsed_base_flow_fy_py_file
    property property_setter_mixins using property_mixins
    property property_constant_setter_mixins using property_mixins
    property user_imports_from_mixins using property_setter_mixins
    property import_abc using constant(["import abc"])
    property import_any using property_setters_exists
    property import_flow_base using constant(["from fy_core.base.flow_base import FlowBase"])
    property import_base_flow using declared_base_flow_name
    property property_mixins_import using property_mixins_and_mixin_import_map
    property method_mixins_import using method_mixins_and_mixin_import_map
    property import_abstract_property_mixins using abstract_property_mixin_and_mixin_import_map
    property import_abstract_method_mixins using abstract_method_mixin_and_mixin_import_map
fy"""

from typing import List, Any, Dict

from fy_core.base.flow_base import FlowBase
from fy_library.domain.mixin_models import (
    MethodMixinModel,
    AbstractMethodModel,
    AbstractPropertyModel,
    PropertyMixinModel,
)
from fy_library.domain.parsed_fy_py_file import ParsedBaseFlowFyPyFile
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.abstract_method_mixins.using_setter import (
    AbstractMethodMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.abstract_property_mixins.using_setter import (
    AbstractPropertyMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.declared_base_flow_name.using_parsed_base_flow_fy_py_file_fy import (
    DeclaredBaseFlowName_UsingParsedBaseFlowFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.import_abc.using_setter import (
    ImportAbc_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.import_flow_base.using_setter import (
    ImportFlowBase_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.imports.import__abstract_property_mixins__using_abstract_property_mixin_and_mixin_import_map__fy import (
    ImportAbstractPropertyMixins_UsingAbstractPropertyMixinAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.imports.import_abstract_method_mixins__using_abstract_method_mixin_and_mixin_import_map__fy import (
    ImportAbstractMethodMixins_UsingAbstractMethodMixinAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.imports.import_any__using_property_setters_exists_fy import (
    ImportAny_UsingPropertySettersExists_PropertyMixin,
)
from fy_library.mixins.property.imports.import_base_flow__using_declared_base_flow_name_fy import (
    ImportBaseFlow_UsingDeclaredBaseFlowName_PropertyMixin,
)

from fy_library.mixins.property.imports.method_mixins_imports__using_method_mixins_and_mixin_import_map_fy import (
    MethodMixinsImport_UsingMethodMixinsAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.imports.property_mixins_imports__using_property_mixins_and_mixin_import_map_fy import (
    PropertyMixinsImport_UsingPropertyMixinsAndMixinImportMap_PropertyMixin,
)
from fy_library.mixins.property.method_mixins.using_setter import (
    MethodMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.mixin_import_map.using_setter import (
    MixinImportMap_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_base_flow_fy_py_file.using_setter import (
    ParsedBaseFlowFyPyFile_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.using_setter import (
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.property_constant_setter_mixins.using_property_mixins_fy import (
    PropertyConstantSetterMixins_UsingPropertyMixins_PropertyMixin,
)
from fy_library.mixins.property.property_mixins.using_setter import (
    PropertyMixins_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.property_setter_mixins.using_property_mixin_fy import (
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
)
from fy_library.mixins.property.user_imports_from_property_mixins.using_property_setter_imports_fy import (
    UserImportsFromMixins_UsingPropertySetterMixins_PropertyMixin,
)

from fy_library.mixins.property.import_generic.using_setter import (
    ImportGeneric_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class BaseFlowImports_Flow(
    # Property Mixins
    PropertyMixins_UsingSetter_PropertyMixin,
    MethodMixins_UsingSetter_PropertyMixin,
    AbstractPropertyMixins_UsingSetter_PropertyMixin,
    AbstractMethodMixins_UsingSetter_PropertyMixin,
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
    MixinImportMap_UsingSetter_PropertyMixin,
    ParsedBaseFlowFyPyFile_UsingSetter_PropertyMixin,
    ImportGeneric_UsingSetter_PropertyMixin,
    DeclaredBaseFlowName_UsingParsedBaseFlowFyPyFile_PropertyMixin,
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
    PropertyConstantSetterMixins_UsingPropertyMixins_PropertyMixin,
    UserImportsFromMixins_UsingPropertySetterMixins_PropertyMixin,
    ImportAbc_UsingSetter_PropertyMixin,
    ImportAny_UsingPropertySettersExists_PropertyMixin,
    ImportFlowBase_UsingSetter_PropertyMixin,
    ImportBaseFlow_UsingDeclaredBaseFlowName_PropertyMixin,
    PropertyMixinsImport_UsingPropertyMixinsAndMixinImportMap_PropertyMixin,
    MethodMixinsImport_UsingMethodMixinsAndMixinImportMap_PropertyMixin,
    ImportAbstractPropertyMixins_UsingAbstractPropertyMixinAndMixinImportMap_PropertyMixin,
    ImportAbstractMethodMixins_UsingAbstractMethodMixinAndMixinImportMap_PropertyMixin,
    # Base
    FlowBase[List[str]],
):
    def __init__(
        self,
        *args: Any,
        property_mixins: List[PropertyMixinModel],
        method_mixins: List[MethodMixinModel],
        abstract_property_mixins: List[AbstractPropertyModel],
        abstract_method_mixins: List[AbstractMethodModel],
        parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile],
        mixin_import_map: Dict[str, str],
        parsed_base_flow_fy_py_file: ParsedBaseFlowFyPyFile,
        **kwargs: Any,
    ):
        self._property_mixins = property_mixins
        self._method_mixins = method_mixins
        self._abstract_property_mixins = abstract_property_mixins
        self._abstract_method_mixins = abstract_method_mixins
        self._parsed_fy_py_files_map_by_key = parsed_fy_py_files_map_by_key
        self._mixin_import_map = mixin_import_map
        self._parsed_base_flow_fy_py_file = parsed_base_flow_fy_py_file
        self._import_generic = ["from typing import Generic"]
        self._import_abc = ["import abc"]
        self._import_flow_base = ["from fy_core.base.flow_base import FlowBase"]
        super().__init__(*args, **kwargs)

    def __call__(self) -> List[str]:
        # fy:end <<<===
        base_flow_import = (
            self._import_flow_base
            if self._import_base_flow == []
            else self._import_base_flow
        )

        import_generic_constant = (
            self._import_generic
            if self._parsed_base_flow_fy_py_file.generics_def != ""
            else []
        )

        return (
            self._user_imports_from_mixins
            + self._import_any
            + self._import_abc
            + import_generic_constant
            + base_flow_import
            + self._property_mixins_import
            + self._method_mixins_import
            + self._import_abstract_property_mixins
            + self._import_abstract_method_mixins
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


property parsed_fy_py_files: List[ParsedFyPyFile] using property_setter_mixins__mapped_to_abstract_property:
    property property_setter_mixins
    property parsed_fy_py_files_map_by_key
fy"""

from functools import cached_property
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)
from fy_library.mixins.property.property_setter_mixins.abc_fy import (
    PropertySetterMixins_PropertyMixin_ABC,
)
from typing import List
import abc

from fy_library.mixins.property.parsed_fy_py_files.abc_fy import (
    ParsedFyPyFiles_PropertyMixin_ABC,
)


# fy:start ===>>>
class ParsedFyPyFiles_UsingPropertySetterMixins_MappedToAbstractProperty_PropertyMixin(
    # Property_mixins
    ParsedFyPyFiles_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    PropertySetterMixins_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _parsed_fy_py_files(self) -> List[ParsedFyPyFile]:
        # fy:end <<<===
        return [
            self._parsed_fy_py_files_map_by_key[
                property_setter.property_name.snake_case
            ]
            for property_setter in self._property_setter_mixins
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.mixins.property.abstract_property_file_split.abc_fy import AbstractPropertyFileSplitModel


property abstract_property_file_split: AbstractPropertyFileSplitModel using abstract_property_regex:
    property fy_py_file_to_parse
    property fy_code
fy"""

import abc
import re
from functools import cached_property
from typing import Final

from fy_library.constants import (
    FY_ENTITY_REGEX_STRING,
    PYTHON_MULTI_ENTITY_REGEX_STRING,
)
from fy_library.mixins.property.abstract_property_file_split.abc_fy import (
    AbstractPropertyFileSplitModel,
    AbstractPropertyFileSplit_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)

_ABSTRACT_PROPERTY_REGEX: Final = re.compile(
    rf"property\s+(?P<abstract_property_name>{FY_ENTITY_REGEX_STRING})"
    rf"(?:\[(?P<generic_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})])?"
    rf":\s*(?P<return_type>{PYTHON_MULTI_ENTITY_REGEX_STRING})\s*$",
    flags=re.DOTALL,
)


# fy:start ===>>>
class AbstractPropertyFileSplit_UsingAbstractPropertyRegex_PropertyMixin(
    # Property_mixins
    AbstractPropertyFileSplit_PropertyMixin_ABC,
    FyCode_PropertyMixin_ABC,
    FyPyFileToParse_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _abstract_property_file_split(self) -> AbstractPropertyFileSplitModel:
        # fy:end <<<===
        abstract_property_file_split = _ABSTRACT_PROPERTY_REGEX.split(self._fy_code)

        assert (
            len(abstract_property_file_split) == 5
        ), f"Abstract property file split length {len(abstract_property_file_split)} in {self._fy_py_file_to_parse}"

        assert (
            abstract_property_file_split[3] is not None
        ), f"Abstract property requires property type in {self._fy_py_file_to_parse}"

        abstract_property_file_split_model = AbstractPropertyFileSplitModel(
            user_imports=abstract_property_file_split[0],
            abstract_property_name=abstract_property_file_split[1],
            generics_def=abstract_property_file_split[2] or "",
            property_type=abstract_property_file_split[3],
        )

        return abstract_property_file_split_model



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property import_abc: List[str] using when_abstract_property_and_abstract_method_exists:
    property abstract_property_mixins
    property abstract_method_mixins
fy"""

from functools import cached_property
from typing import List


from fy_library.mixins.property.abstract_method_mixins.abc_fy import (
    AbstractMethodMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.abstract_property_mixins.abc_fy import (
    AbstractPropertyMixins_PropertyMixin_ABC,
)
import abc


from fy_library.mixins.property.import_abc.abc_fy import ImportAbc_PropertyMixin_ABC


# fy:start ===>>>
class ImportAbc_UsingWhenAbstractPropertyAndAbstractMethodExists_PropertyMixin(
    # Property_mixins
    AbstractMethodMixins_PropertyMixin_ABC,
    AbstractPropertyMixins_PropertyMixin_ABC,
    ImportAbc_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _import_abc(self) -> List[str]:
        # fy:end <<<===
        return (
            ["import abc"]
            if (self._abstract_property_mixins or self._abstract_method_mixins)
            else []
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.fy_py_template_models import BaseFlowTemplateModel


flow create_base_flow_template_model__using_parsed_fy_py_file_and_property_setters_template_models -> BaseFlowTemplateModel:
    property parsed_fy_py_file using setter
    property parsed_fy_py_files_map_by_key using setter
    property property_mixins using template_model_properties
    property property_setter_mixins using property_mixins
    property parsed_fy_py_files using property_setter_mixins__mapped_to_abstract_property
    property abstract_entities_ordering_index using setter
    property abstract_mixins using parsed_base_flow_fy_py_file
    property mro_ordered_abstract_mixins using abstract_mixins_and_ordered_abstract_entities
fy"""

from typing import Any, List
from typing import Dict

from fy_core.base.flow_base import FlowBase

from fy_library.domain.annotation_object import AnnotationKind
from fy_library.domain.fy_py_template_models import (
    BaseFlowTemplateModel,
)
from fy_library.domain.mixin_models import BaseMixinModel, MixinModelKind
from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFile,
    ParsedBaseFlowFyPyFile,
)
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.abstract_mixins.using_parsed_base_flow_fy_py_file_fy import (
    AbstractMixins_UsingParsedBaseFlowFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.mro_ordered_abstract_mixins.new_parsed_fy_py_file_and_ordered_abstract_entities_fy import (
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
)
from fy_library.mixins.property.ordered_abstract_entities.using_setter import (
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files.using_property_setter_mixins__mapped_to_abstract_property__fy import (
    ParsedFyPyFiles_UsingPropertySetterMixins_MappedToAbstractProperty_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.using_setter import (
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.property_mixins.using_template_model_properties_fy import (
    PropertyMixins_UsingTemplateModelProperties_PropertyMixin,
)
from fy_library.mixins.property.property_setter_mixins.using_property_mixin_fy import (
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
)


# fy:start ===>>>
class CreateBaseFlowTemplateModel_UsingParsedFyPyFileAndPropertySettersTemplateModels_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    ParsedFyPyFilesMapByKey_UsingSetter_PropertyMixin,
    PropertyMixins_UsingTemplateModelProperties_PropertyMixin,
    PropertySetterMixins_UsingPropertyMixins_PropertyMixin,
    ParsedFyPyFiles_UsingPropertySetterMixins_MappedToAbstractProperty_PropertyMixin,
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
    AbstractMixins_UsingParsedBaseFlowFyPyFile_PropertyMixin,
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
    # Base
    FlowBase[BaseFlowTemplateModel],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        parsed_fy_py_files_map_by_key: Dict[str, ParsedFyPyFile],
        abstract_entities_ordering_index: Dict[str, int],
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        self._parsed_fy_py_files_map_by_key = parsed_fy_py_files_map_by_key
        self._abstract_entities_ordering_index = abstract_entities_ordering_index
        super().__init__(*args, **kwargs)

    def __call__(self) -> BaseFlowTemplateModel:
        # fy:end <<<===
        parsed_base_flow_fy_py_file = self._parsed_fy_py_file
        assert isinstance(parsed_base_flow_fy_py_file, ParsedBaseFlowFyPyFile)

        callable_annotation = any(
            AnnotationKind.CALLABLE == annotation.kind
            for annotation in parsed_base_flow_fy_py_file.annotations
        )

        abstract_method_mixins: List[BaseMixinModel] = [
            abstract_property
            for abstract_property in self._mro_ordered_abstract_mixins
            if abstract_property.kind == MixinModelKind.ABSTRACT_METHOD
        ]
        abstract_property_mixins: List[BaseMixinModel] = [
            abstract_method
            for abstract_method in self._mro_ordered_abstract_mixins
            if abstract_method.kind == MixinModelKind.ABSTRACT_PROPERTY
        ]

        return BaseFlowTemplateModel(
            python_class_name=parsed_base_flow_fy_py_file.python_class_name,
            base_flow_name=parsed_base_flow_fy_py_file.base_flow_name,
            generics_def=parsed_base_flow_fy_py_file.generics_def,
            declared_base_flow=PythonEntityName.from_pascal_case(
                f"{parsed_base_flow_fy_py_file.declared_base_flow}_BaseFlow"
            )
            if parsed_base_flow_fy_py_file.declared_base_flow != ""
            else None,
            callable_annotation=callable_annotation,
            return_type=parsed_base_flow_fy_py_file.return_type,
            properties=parsed_base_flow_fy_py_file.properties,
            methods=parsed_base_flow_fy_py_file.methods,
            abstract_property_mixins=abstract_property_mixins,
            abstract_method_mixins=abstract_method_mixins,
            property_setters=self._parsed_fy_py_files,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
def entity_key(
    mixin_name__snake_case: str, mixin_implementation_name__snake_case: str
) -> str:
    return f"{mixin_name__snake_case}.{mixin_implementation_name__snake_case}"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedBaseFlowFyPyFile


property parsed_base_flow_fy_py_file: ParsedBaseFlowFyPyFile using parsed_fy_py_file:
    property pre_fy_code
    property fy_code
    property pre_marker_file_content
    property post_marker_file_content
    property fy_py_file_to_parse
    property base_flow_file_split
    property included_mixins
fy"""

import abc
from functools import cached_property

from fy_library.domain.fy_py_template_models import TemporaryBaseTemplateModel
from fy_library.domain.parsed_fy_py_file import ParsedBaseFlowFyPyFile
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.base_flow_file_split.abc_fy import (
    BaseFlowFileSplit_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)
from fy_library.mixins.property.included_mixins.abc_fy import (
    IncludedMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.post_marker_file_content.abc_fy import (
    PostMarkerFileContent_PropertyMixin_ABC,
)
from fy_library.mixins.property.pre_fy_code.abc_fy import (
    PreFyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.pre_marker_file_content.abc_fy import (
    PreMarkerFileContent_PropertyMixin_ABC,
)

from fy_library.mixins.property.parsed_base_flow_fy_py_file.abc_fy import (
    ParsedBaseFlowFyPyFile_PropertyMixin_ABC,
)


# fy:start ===>>>
class ParsedBaseFlowFyPyFile_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    BaseFlowFileSplit_PropertyMixin_ABC,
    FyCode_PropertyMixin_ABC,
    FyPyFileToParse_PropertyMixin_ABC,
    IncludedMixins_PropertyMixin_ABC,
    ParsedBaseFlowFyPyFile_PropertyMixin_ABC,
    PostMarkerFileContent_PropertyMixin_ABC,
    PreFyCode_PropertyMixin_ABC,
    PreMarkerFileContent_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _parsed_base_flow_fy_py_file(self) -> ParsedBaseFlowFyPyFile:
        # fy:end <<<===

        base_flow_name = PythonEntityName.from_snake_case(
            self._base_flow_file_split.base_flow_name
        )

        parsed_fy_py_file = ParsedBaseFlowFyPyFile(
            pre_fy_code=self._pre_fy_code,
            fy_code=self._fy_code,
            pre_marker_file_content=self._pre_marker_file_content,
            post_marker_file_content=self._post_marker_file_content,
            file_path=self._fy_py_file_to_parse,
            user_imports=self._base_flow_file_split.user_imports,
            annotations=self._base_flow_file_split.annotations,
            base_flow_name=base_flow_name,
            generics_def=self._base_flow_file_split.generics_def,
            declared_base_flow=self._base_flow_file_split.declared_base_flow,
            return_type=self._base_flow_file_split.return_type,
            properties=self._included_mixins.property_mixins,
            methods=self._included_mixins.method_mixins,
            abstract_property_mixins=self._included_mixins.abstract_property_mixins,
            abstract_method_mixins=self._included_mixins.abstract_method_mixins,
            python_class_name=PythonEntityName.from_pascal_case(
                f"{base_flow_name.pascal_case}_BaseFlow"
            ),
            # TODO: remove after template_model removed from the base class.
            template_model=TemporaryBaseTemplateModel(
                python_class_name=PythonEntityName.from_pascal_case(
                    f"{base_flow_name.pascal_case}_BaseFlow"
                ),
                entity_key_value=base_flow_name.snake_case,
            ),
        )

        return parsed_fy_py_file



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List
from fy_library.domain.parsed_fy_py_file import PropertySetterFyPyFile


property required_property_setters_fy_py: List[PropertySetterFyPyFile] using parsed_fy_py_files:
    property parsed_fy_py_files
    property parsed_fy_py_files_map_by_key
    property required_setters
fy"""

import abc
from functools import cached_property
from typing import List

from fy_library.domain.parsed_fy_py_file import (
    PropertySetterFyPyFile,
)
from fy_library.mixins.property.parsed_fy_py_files.abc_fy import (
    ParsedFyPyFiles_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_files_map_by_key.abc_fy import (
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
)
from fy_library.mixins.property.required_property_setters_fy_py.abc_fy import (
    RequiredPropertySettersFyPy_PropertyMixin_ABC,
)
from fy_library.mixins.property.required_setters.abc_fy import (
    RequiredSetters_PropertyMixin_ABC,
)


# fy:start ===>>>
class RequiredPropertySettersFyPy_UsingParsedFyPyFiles_PropertyMixin(
    # Property_mixins
    ParsedFyPyFiles_PropertyMixin_ABC,
    ParsedFyPyFilesMapByKey_PropertyMixin_ABC,
    RequiredPropertySettersFyPy_PropertyMixin_ABC,
    RequiredSetters_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _required_property_setters_fy_py(self) -> List[PropertySetterFyPyFile]:
        # fy:end <<<===
        required_setters = self._required_setters
        return list(required_setters.values())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow generate_and_save_fy_py_file__using_required_property_setters -> None:
    property parsed_fy_py_file using setter
    property import_generic using constant(["from typing import Generic"])
    property jinja2_template_file_name using property_setter_constant
    property template_model using parsed_fy_py_file__for_setter
    property generated_fy_py_code using jinja2_templates
    property fy_py_file_content using required_property_setter
    method generate_and_save_fy_py_code using parsed_fy_py_file__and__fy_py_file_content
fy"""

from typing import Any

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.method.generate_and_save_fy_py_code.using_parsed_fy_py_file__and__fy_py_file_content_fy import (
    GenerateAndSaveFyPyCode_UsingParsedFyPyFile_And_FyPyFileContent_MethodMixin,
)
from fy_library.mixins.property.fy_py_file_content.using_required_property_setter_fy import (
    FyPyFileContent_UsingRequiredPropertySetter_PropertyMixin,
)
from fy_library.mixins.property.generated_fy_py_code.using_jinja2_templates_fy import (
    GeneratedFyPyCode_UsingJinja2Templates_PropertyMixin,
)
from fy_library.mixins.property.jinja2_template_file_name.using_property_setter_constant_fy import (
    Jinja2TemplateFileName_UsingPropertySetterConstant_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.template_model.using_parsed_fy_py_file_fy__for_setter_fy import (
    TemplateModel_UsingParsedFyPyFile_ForSetter_PropertyMixin,
)

from fy_library.mixins.property.import_generic.using_setter import (
    ImportGeneric_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class GenerateAndSaveFyPyFile_UsingRequiredPropertySetters_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    ImportGeneric_UsingSetter_PropertyMixin,
    Jinja2TemplateFileName_UsingPropertySetterConstant_PropertyMixin,
    TemplateModel_UsingParsedFyPyFile_ForSetter_PropertyMixin,
    GeneratedFyPyCode_UsingJinja2Templates_PropertyMixin,
    FyPyFileContent_UsingRequiredPropertySetter_PropertyMixin,
    # Method Mixins
    GenerateAndSaveFyPyCode_UsingParsedFyPyFile_And_FyPyFileContent_MethodMixin,
    # Base
    FlowBase[None],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        self._import_generic = ["from typing import Generic"]
        super().__init__(*args, **kwargs)

    def __call__(self) -> None:
        # fy:end <<<===
        self._generate_and_save_fy_py_code()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.fy_py_template_models import BaseTemplateModel


property template_model: BaseTemplateModel using parsed_fy_py_file__for_setter:
    property parsed_fy_py_file
fy"""

import abc
from functools import cached_property

from fy_library.domain.fy_py_template_models import BaseTemplateModel
from fy_library.flows.create_template_model_using_parsed_fy_py_file.property_setter_template_model_fy import (
    CreatePropertySetterTemplateModel_UsingParsedFyPyFile_Flow,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)

from fy_library.mixins.property.template_model.abc_fy import (
    TemplateModel_PropertyMixin_ABC,
)


# fy:start ===>>>
class TemplateModel_UsingParsedFyPyFile_ForSetter_PropertyMixin(
    # Property_mixins
    ParsedFyPyFile_PropertyMixin_ABC,
    TemplateModel_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _template_model(self) -> BaseTemplateModel:
        # fy:end <<<===
        return CreatePropertySetterTemplateModel_UsingParsedFyPyFile_Flow(
            parsed_fy_py_file=self._parsed_fy_py_file,
        )()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedMethodFyPyFile


property parsed_method_fy_py_file: ParsedMethodFyPyFile using parsed_fy_py_file:
    property pre_fy_code
    property fy_code
    property pre_marker_file_content
    property post_marker_file_content
    property fy_py_file_to_parse
    property method_file_split
    property included_mixins
fy"""

import abc
from functools import cached_property

from fy_library.domain.fy_py_template_models import (
    TemporaryBaseTemplateModel,
)
from fy_library.domain.parsed_fy_py_file import ParsedMethodFyPyFile
from fy_library.domain.entity_key import entity_key
from fy_library.domain.python_entity_name import PythonEntityName
from fy_library.mixins.property.fy_code.abc_fy import (
    FyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.fy_py_file_to_parse.abc_fy import (
    FyPyFileToParse_PropertyMixin_ABC,
)
from fy_library.mixins.property.included_mixins.abc_fy import (
    IncludedMixins_PropertyMixin_ABC,
)
from fy_library.mixins.property.method_file_split.abc_fy import (
    MethodFileSplit_PropertyMixin_ABC,
)
from fy_library.mixins.property.post_marker_file_content.abc_fy import (
    PostMarkerFileContent_PropertyMixin_ABC,
)
from fy_library.mixins.property.pre_fy_code.abc_fy import (
    PreFyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.pre_marker_file_content.abc_fy import (
    PreMarkerFileContent_PropertyMixin_ABC,
)

from fy_library.mixins.property.parsed_method_fy_py_file.abc_fy import (
    ParsedMethodFyPyFile_PropertyMixin_ABC,
)


# fy:start ===>>>
class ParsedMethodFyPyFile_UsingParsedFyPyFile_PropertyMixin(
    # Property_mixins
    FyCode_PropertyMixin_ABC,
    FyPyFileToParse_PropertyMixin_ABC,
    IncludedMixins_PropertyMixin_ABC,
    MethodFileSplit_PropertyMixin_ABC,
    ParsedMethodFyPyFile_PropertyMixin_ABC,
    PostMarkerFileContent_PropertyMixin_ABC,
    PreFyCode_PropertyMixin_ABC,
    PreMarkerFileContent_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _parsed_method_fy_py_file(self) -> ParsedMethodFyPyFile:
        # fy:end <<<===

        method_name = PythonEntityName.from_snake_case(
            self._method_file_split.method_name
        )
        implementation_name = PythonEntityName.from_snake_case(
            self._method_file_split.implementation_name
        )

        parsed_fy_py_file = ParsedMethodFyPyFile(
            pre_fy_code=self._pre_fy_code,
            fy_code=self._fy_code,
            pre_marker_file_content=self._pre_marker_file_content,
            post_marker_file_content=self._post_marker_file_content,
            file_path=self._fy_py_file_to_parse,
            user_imports=self._method_file_split.user_imports,
            method_name=method_name,
            abstract_property_mixins=self._included_mixins.abstract_property_mixins,
            abstract_method_mixins=self._included_mixins.abstract_method_mixins,
            generics_def=self._method_file_split.generics_def,
            arguments=self._method_file_split.arguments,
            return_type=self._method_file_split.return_type,
            implementation_name=implementation_name,
            python_class_name=PythonEntityName.from_pascal_case(
                f"{method_name.pascal_case}_Using{implementation_name.pascal_case}_MethodMixin"
            ),
            template_model=TemporaryBaseTemplateModel(
                python_class_name=PythonEntityName.from_pascal_case(
                    f"{method_name.pascal_case}_Using{implementation_name.pascal_case}_MethodMixin"
                ),
                entity_key_value=entity_key(
                    mixin_name__snake_case=method_name.snake_case,
                    mixin_implementation_name__snake_case=implementation_name.snake_case,
                ),
            ),
        )

        return parsed_fy_py_file



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile


flow parse_flow_fy_code -> ParsedFyPyFile:
    property pre_fy_code using setter
    property fy_code using setter
    property pre_marker_file_content using setter
    property post_marker_file_content using setter
    property fy_py_file_to_parse using setter
    property flow_file_split using flow_regex
    property mixin_lines using flow_file_split
    property mixin_models using mixin_lines
    property included_mixins using mixin_models
    property parsed_flow_fy_py_file using parsed_fy_py_file
fy"""

from pathlib import Path
from typing import Any

from fy_core.base.flow_base import FlowBase
from fy_library.domain.parsed_fy_py_file import ParsedFyPyFile
from fy_library.mixins.property.flow_file_split.using_flow_regex_fy import (
    FlowFileSplit_UsingFlowRegex_PropertyMixin,
)
from fy_library.mixins.property.fy_code.using_setter import (
    FyCode_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.fy_py_file_to_parse.using_setter import (
    FyPyFileToParse_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.included_mixins.using_mixin_lines_fy import (
    IncludedMixins_UsingMixinModels_PropertyMixin,
)
from fy_library.mixins.property.mixin_lines.using_flow_file_split_fy import (
    MixinLines_UsingFlowFileSplit_PropertyMixin,
)
from fy_library.mixins.property.mixin_models.using_mixin_lines_fy import (
    MixinModels_UsingMixinLines_PropertyMixin,
)
from fy_library.mixins.property.parsed_flow_fy_py_file.using_parsed_fy_py_file_fy import (
    ParsedFlowFyPyFile_UsingParsedFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.post_marker_file_content.using_setter import (
    PostMarkerFileContent_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.pre_fy_code.using_setter import (
    PreFyCode_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.pre_marker_file_content.using_setter import (
    PreMarkerFileContent_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class ParseFlowFyCode_Flow(
    # Property Mixins
    PreFyCode_UsingSetter_PropertyMixin,
    FyCode_UsingSetter_PropertyMixin,
    PreMarkerFileContent_UsingSetter_PropertyMixin,
    PostMarkerFileContent_UsingSetter_PropertyMixin,
    FyPyFileToParse_UsingSetter_PropertyMixin,
    FlowFileSplit_UsingFlowRegex_PropertyMixin,
    MixinLines_UsingFlowFileSplit_PropertyMixin,
    MixinModels_UsingMixinLines_PropertyMixin,
    IncludedMixins_UsingMixinModels_PropertyMixin,
    ParsedFlowFyPyFile_UsingParsedFyPyFile_PropertyMixin,
    # Base
    FlowBase[ParsedFyPyFile],
):
    def __init__(
        self,
        *args: Any,
        pre_fy_code: str,
        fy_code: str,
        pre_marker_file_content: str,
        post_marker_file_content: str,
        fy_py_file_to_parse: Path,
        **kwargs: Any,
    ):
        self._pre_fy_code = pre_fy_code
        self._fy_code = fy_code
        self._pre_marker_file_content = pre_marker_file_content
        self._post_marker_file_content = post_marker_file_content
        self._fy_py_file_to_parse = fy_py_file_to_parse
        super().__init__(*args, **kwargs)

    def __call__(self) -> ParsedFyPyFile:
        # fy:end <<<===
        assert len(self._included_mixins.abstract_method_mixins) == 0, (
            f"Flow {self._fy_py_file_to_parse} cannot include "
            f"other abstract method implementations."
        )
        assert len(self._included_mixins.abstract_property_mixins) == 0, (
            f"Flow {self._fy_py_file_to_parse} cannot include "
            f"other abstract property implementations."
        )
        return self._parsed_flow_fy_py_file



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
flow create_method_template_model__using_parsed_fy_py_file -> MethodTemplateModel:
    property parsed_fy_py_file using setter
    property abstract_entities_ordering_index using setter
    property abstract_mixins using parsed_method_fy_py_file
    property mro_ordered_abstract_mixins using abstract_mixins_and_ordered_abstract_entities
fy"""

from typing import Any, Dict, List

from fy_core.base.flow_base import FlowBase

from fy_library.domain.fy_py_template_models import MethodTemplateModel
from fy_library.domain.mixin_models import BaseMixinModel, MixinModelKind
from fy_library.domain.parsed_fy_py_file import (
    ParsedFyPyFile,
    ParsedMethodFyPyFile,
)
from fy_library.mixins.property.abstract_mixins.using_parsed_method_fy_py_file_with_filtered_method_mixins_fy import (
    AbstractMixins_UsingParsedMethodFyPyFile_PropertyMixin,
)
from fy_library.mixins.property.mro_ordered_abstract_mixins.new_parsed_fy_py_file_and_ordered_abstract_entities_fy import (
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
)
from fy_library.mixins.property.ordered_abstract_entities.using_setter import (
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
)
from fy_library.mixins.property.parsed_fy_py_file.using_setter import (
    ParsedFyPyFile_UsingSetter_PropertyMixin,
)


# fy:start ===>>>
class CreateMethodTemplateModel_UsingParsedFyPyFile_Flow(
    # Property Mixins
    ParsedFyPyFile_UsingSetter_PropertyMixin,
    AbstractEntitiesOrderingIndex_UsingSetter_PropertyMixin,
    AbstractMixins_UsingParsedMethodFyPyFile_PropertyMixin,
    MroOrderedAbstractMixins_UsingAbstractMixinsAndOrderedAbstractEntities_PropertyMixin,
    # Base
    FlowBase[MethodTemplateModel],
):
    def __init__(
        self,
        *args: Any,
        parsed_fy_py_file: ParsedFyPyFile,
        abstract_entities_ordering_index: Dict[str, int],
        **kwargs: Any,
    ):
        self._parsed_fy_py_file = parsed_fy_py_file
        self._abstract_entities_ordering_index = abstract_entities_ordering_index
        super().__init__(*args, **kwargs)

    def __call__(self) -> MethodTemplateModel:
        # fy:end <<<===
        parsed_method_fy_py_file = self._parsed_fy_py_file
        assert isinstance(parsed_method_fy_py_file, ParsedMethodFyPyFile)

        abstract_method_mixins: List[BaseMixinModel] = [
            abstract_property
            for abstract_property in self._mro_ordered_abstract_mixins
            if abstract_property.kind == MixinModelKind.ABSTRACT_METHOD
        ]
        abstract_property_mixins: List[BaseMixinModel] = [
            abstract_method
            for abstract_method in self._mro_ordered_abstract_mixins
            if abstract_method.kind == MixinModelKind.ABSTRACT_PROPERTY
        ]

        return MethodTemplateModel(
            python_class_name=parsed_method_fy_py_file.python_class_name,
            method_name=parsed_method_fy_py_file.method_name,
            abstract_method_mixins=abstract_method_mixins,
            abstract_property_mixins=abstract_property_mixins,
            generics_def=parsed_method_fy_py_file.generics_def,
            arguments=parsed_method_fy_py_file.arguments,
            implementation_name=parsed_method_fy_py_file.implementation_name,
            return_type=parsed_method_fy_py_file.return_type,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property mixin_imports_code: str using filtered_mixin_imports:
    property filtered_mixin_imports
fy"""

import abc
from functools import cached_property

from fy_library.mixins.property.filtered_mixin_imports.abc_fy import (
    FilteredMixinImports_PropertyMixin_ABC,
)

from fy_library.mixins.property.mixin_imports_code.abc_fy import (
    MixinImportsCode_PropertyMixin_ABC,
)


# fy:start ===>>>
class MixinImportsCode_UsingFilteredMixinImports_PropertyMixin(
    # Property_mixins
    FilteredMixinImports_PropertyMixin_ABC,
    MixinImportsCode_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _mixin_imports_code(self) -> str:
        # fy:end <<<===
        return "\n".join(
            self._filtered_mixin_imports
            + ([""] if self._filtered_mixin_imports else [])
        )



# fy:start ===>>>
from typing import List


class ImportFlowBase_UsingSetter_PropertyMixin:
    @property
    def _import_flow_base(self) -> List[str]:
        return self.__import_flow_base

    @_import_flow_base.setter
    def _import_flow_base(self, import_flow_base: List[str]) -> None:
        self.__import_flow_base = import_flow_base


# fy:end <<<===



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
property generated_fy_py_code: str using jinja2_templates:
    property template_model
    property jinja2_template_file_name
fy"""

import abc
import pathlib
from functools import cached_property
from typing import Final

import jinja2
from jinja2 import Environment, FileSystemLoader

from fy_library.mixins.property.generated_fy_py_code.abc_fy import (
    GeneratedFyPyCode_PropertyMixin_ABC,
)
from fy_library.mixins.property.jinja2_template_file_name.abc_fy import (
    Jinja2TemplateFileName_PropertyMixin_ABC,
)
from fy_library.mixins.property.template_model.abc_fy import (
    TemplateModel_PropertyMixin_ABC,
)

_TEMPLATES_PATH: Final = pathlib.Path(__file__).parent / "jinja2_templates"
_JINJA2_TEMPLATE_ENVIRONMENT: Final = Environment(
    loader=FileSystemLoader(_TEMPLATES_PATH),
    undefined=jinja2.StrictUndefined,
)


# fy:start ===>>>
class GeneratedFyPyCode_UsingJinja2Templates_PropertyMixin(
    # Property_mixins
    GeneratedFyPyCode_PropertyMixin_ABC,
    Jinja2TemplateFileName_PropertyMixin_ABC,
    TemplateModel_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _generated_fy_py_code(self) -> str:
        # fy:end <<<===
        template = _JINJA2_TEMPLATE_ENVIRONMENT.get_template(
            self._jinja2_template_file_name
        )
        content = template.render(self._template_model.model_dump())
        return content



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""fy
from typing import List


property filtered_mixin_imports: List[str] using remove_existing_imports:
    property mixin_imports
    property parsed_fy_py_file
fy"""

import abc
import re
from functools import cached_property
from typing import List, Final

from fy_library.mixins.property.filtered_mixin_imports.abc_fy import (
    FilteredMixinImports_PropertyMixin_ABC,
)
from fy_library.mixins.property.mixin_imports.abc_fy import (
    MixinImports_PropertyMixin_ABC,
)
from fy_library.mixins.property.parsed_fy_py_file.abc_fy import (
    ParsedFyPyFile_PropertyMixin_ABC,
)

_IMPORT_REGEX: Final = re.compile(
    r"^(?P<from>from [\w.]+) import (?P<from_classes>.*)\s*$|^(?P<import>import [\w.]+)$",
    flags=re.DOTALL,
)


# fy:start ===>>>
class FilteredMixinImports_UsingRemoveExistingImports_PropertyMixin(
    # Property_mixins
    FilteredMixinImports_PropertyMixin_ABC,
    MixinImports_PropertyMixin_ABC,
    ParsedFyPyFile_PropertyMixin_ABC,
    abc.ABC,
):
    @cached_property
    def _filtered_mixin_imports(self) -> List[str]:
        # fy:end <<<===
        pre_marker_imports = set(
            self.__collect_import_classes(
                import_lines=self._parsed_fy_py_file.pre_marker_file_content.split(
                    "\n"
                ),
                pre_marker_imports=set(),
            )
        )

        mixin_imports_result = self.__collect_import_classes(
            import_lines=self._mixin_imports, pre_marker_imports=pre_marker_imports
        )

        user_imports_results = self.__collect_import_classes(
            import_lines=self._parsed_fy_py_file.user_imports.split("\n"),
            pre_marker_imports=pre_marker_imports,
        )

        def import_sort_key(import_line: str) -> str:
            import_line_split = import_line.split(" ")
            return "".join(
                ["0" if import_line_split[0] == "import" else "1"]
                + import_line_split[1:]
            )

        return self.__format_imports(
            sorted(list(mixin_imports_result), key=lambda i: import_sort_key(i))
            + sorted(list(user_imports_results), key=lambda i: import_sort_key(i))
        )

    def __format_imports(self, imports: list[str]) -> list[str]:
        result_imports: list[str] = []
        for import_line in imports:
            if import_line.startswith("import ") or len(import_line) <= 88:
                result_imports.append(import_line)
                continue
            import_line_parts = import_line.split(" ")
            result_imports.append(
                f"{' '.join(import_line_parts[:-1])} (\n{' ' * 4}{import_line_parts[-1]},\n)"
            )

        return result_imports

    def __collect_import_classes(
        self,
        import_lines: list[str],
        pre_marker_imports: set[str],
    ) -> list[str]:
        result_imports: list[str] = []

        def import_line_generate_and_append(import_classes: str, from_package: str):
            for import_class in import_classes.split(","):
                import_class = import_class.strip()
                if import_class != "":
                    generated_import_line = f"{from_package} import {import_class}"
                    if generated_import_line not in pre_marker_imports:
                        result_imports.append(generated_import_line)

        collecting_from_import: str | None = None

        for pre_marker_line in import_lines:
            if collecting_from_import is not None:
                end_of_imports = pre_marker_line.strip().endswith(")")
                if end_of_imports:
                    pre_marker_line = pre_marker_line.strip()[:-1]
                import_line_generate_and_append(
                    import_classes=pre_marker_line, from_package=collecting_from_import
                )
                if end_of_imports:
                    collecting_from_import = None

                continue

            import_regex_result = _IMPORT_REGEX.search(pre_marker_line)
            if import_regex_result is None:
                continue

            import_line = import_regex_result.group("import")
            if import_line is not None:
                if import_line not in pre_marker_imports:
                    result_imports.append(import_line)
                continue

            from_import = import_regex_result.group("from")
            assert from_import is not None
            from_classes = import_regex_result.group("from_classes")
            assert from_classes is not None

            start_of_collection = from_classes.strip().startswith("(")
            if start_of_collection:
                from_classes = from_classes.strip()[1:]
            import_line_generate_and_append(
                import_classes=from_classes, from_package=from_import
            )
            if start_of_collection:
                collecting_from_import = from_import
            continue

        return result_imports
